#include <stdio.h>
#include <stdlib.h>
#include <dos.h>
#include <i86.h>
#include <string.h>

int try_ps2_load(void);
void tsr_end(void);

extern unsigned char __based(__segname("_CODE")) acceleration;
extern unsigned char __based(__segname("_CODE")) resolution;
extern unsigned char __based(__segname("_CODE")) samplerate;

unsigned char serial = 0;

void load_settings(void)
{
  char buff[100];
  const char accstr[] = "ACCELERATION";
  const char resstr[] = "RESOLUTION";
  const char srstr[] = "SAMPLERATE";
  const unsigned char samplerates[] = 
  {
    10,
    20,
    40,
    60,
    80,
    100,
    200
  };
  unsigned n;
  FILE *fp = fopen("mouse.ini", "r");
  if(!fp) return;

  while(fgets(buff, sizeof(buff), fp))
  {
    if(!memcmp(buff, accstr, sizeof(accstr)-1))
    {
      acceleration = atoi(&buff[sizeof(accstr)]);
      printf("Acceleration: %u\n", acceleration);
      acceleration = 8 - acceleration;
    }
    else if(!serial && !memcmp(buff, resstr, sizeof(resstr)-1))
    {
      resolution = atoi(&buff[sizeof(resstr)]);
      printf("Resolution: %u (%u counts per mm)\n", resolution, 1<<resolution);
    }
    else if(!serial && !memcmp(buff, srstr, sizeof(srstr)-1))
    {
      samplerate = atoi(&buff[sizeof(srstr)]);
      for(n=0;n+1<sizeof(samplerates)/sizeof(unsigned char)&&samplerate>samplerates[n];n++);
      printf("Sample rate: %u reports per second\n", samplerates[n]);
      samplerate = n;
    }
  }

  fclose(fp);
}

void free_mem(void)
{
  unsigned n;

  for(n=0;!_dos_close(n);n++);

  _dos_freemem(*(unsigned *)0x2C); // free environment, data segment should 
                                   // already point to PSP...
}

void __interrupt int33_handler();

#include "serial.c"

int main(void)
{
  puts("Mouse driver v0.11 (c) Sami Tikkanen 2024");

  if(try_ps2_load())
  {
    puts("No PS/2 mouse detected.");
    if(detect_serial_mouse())
      return 1;
    else serial = 1;
  }

  load_settings();

  _dos_setvect(0x33, int33_handler);

  free_mem();

  _dos_keep(0, 0x10 + ( (FP_OFF(tsr_end)+0xF) >> 4 ));
}
