.8086

.MODEL small

public _acceleration
public _resolution
public _samplerate
public try_ps2_load_
extern load_settings_:BYTE
extern free_mem_:BYTE
public tsr_end_
public int33_handler_
public sermouse_handler_
public _sermouse_port
public _old_serport_handler

LEFT_BUTTON EQU 0x01
CENTER_BUTTON EQU 0x04
RIGHT_BUTTON EQU 0x02
MOUSE_ENABLED EQU 0x08

DIR_LEFT EQU 0x10
DIR_RIGHT EQU 0x20
DIR_UP EQU 0x40
DIR_DOWN EQU 0x80

KEYPAD_LEFT EQU 0x4B
KEYPAD_RIGHT EQU 0x4D
KEYPAD_UP EQU 0x48
KEYPAD_DOWN EQU 0x50

KEY_CTRL EQU 0x1D
KEY_ENABLEMOUSE EQU KEY_CTRL

KEYPAD_INS EQU 0x52
MOUSE_LEFT EQU KEYPAD_INS

KEYPAD_ENTER EQU 0x1C
MOUSE_RIGHT EQU KEYPAD_ENTER

KEYPAD_PLUS EQU 0x4e
MOUSE_CENTER EQU KEYPAD_PLUS


.code

ORG 0x100

; global variables
mouse_button_status: db 0
mouse_x_speed: dw 0
mouse_y_speed: dw 0
cursor_row: dw 100
cursor_col: dw 160
mouse_dx: dw 0
mouse_dy: dw 0

_acceleration: db 1

accelerate:
  push cx

  mov cl, byte ptr cs:_acceleration
  mov ch, al

  test ah, ah
  jz accelerate_pos

  not ch
  inc ch

  accelerate_pos:

  shr ch, cl 

  mov cl, ch

  shl ax, cl

  pop cx
ret

ps2mouse_handler_ proc far
  push bp
  mov bp, sp
  push ax
  push bx

  mov bx, [bp+12] ; status
  mov cs:mouse_button_status, bl
  and cs:mouse_button_status, 0x7


;  mov ax, [bp+6] ; Z data

  mov ax, [bp+8] ; Y data
  test bl, 0x20
  jz y_positive
  dec ah

  y_positive:
  call accelerate
  sub cs:mouse_dy, ax
  sub cs:cursor_row, ax
 
  x_data:
  mov ax, [bp+10] ; X data
  test bl, 0x10
  jz x_positive
  dec ah

  x_positive:
  call accelerate
  add cs:mouse_dx, ax
  add cs:cursor_col, ax
  
  pop bx
  pop ax
  pop bp
  retf
ps2mouse_handler_ endp

_sermouse_port: dw 0
upperbits_save: db 0
sercharcount: db 0

sermouse_handler_ proc far
  push ax
  push dx

  mov dx, cs:_sermouse_port
  add dx, 2
  in al, dx
  test al, 0x01 ; was the interrupt caused by this serial port?
  jnz sermouse_handler_next

  push cx

  sub dx, 2
  in al, dx ; read character

  ; is this first character?
  test al, 0x40
  jz sermouse_notfirst
  mov cs:sercharcount, 1

  ; left button pressed
  mov ah, al
  mov cl, 5
  shr al, cl
  and al, 1
  mov cs:mouse_button_status, al

  ; right button pressed
  mov al, ah
  mov cl, 3
  shr al, cl
  and al, 2
  or cs:mouse_button_status, al

  ; upper 2 bits of x and y
  mov cs:upperbits_save, ah
  jmp sermouse_handler_end

  sermouse_notfirst:
  xor ch, ch
  ; was this the second character?
  cmp cs:sercharcount, 1
  jne sermouse_notsecond
  inc cs:sercharcount

  ; lower 6 bits of horizontal motion
  mov ah, cs:upperbits_save
  mov cl, 6
  shl ah, cl
  or al, ah
  ; is the number positive?
  jns hor_notsign
  dec ch

  hor_notsign:
  mov ah, ch
  call accelerate
  add cs:mouse_dx, ax
  add cs:cursor_col, ax

  jmp sermouse_handler_end

  sermouse_notsecond:
  ; was this the third character?
  cmp cs:sercharcount, 2
  jne sermouse_notthird
  inc cs:sercharcount

  ; lower 6 bits of vertical motion
  mov ah, cs:upperbits_save
  mov cl, 4
  shl ah, cl
  and ah, 0xC0
  or al, ah
  ; is the number positive?
  jns ver_notsign
  dec ch

  ver_notsign:
  mov ah, ch
  call accelerate
  add cs:mouse_dy, ax
  add cs:cursor_row, ax

  jmp sermouse_handler_end

  sermouse_notthird:
  ; was this the fourth character?
  cmp cs:sercharcount, 3
  jne sermouse_notfourth
  inc cs:sercharcount

  ; is the middle button pressed?
  mov cl, 3
  shr al, cl
  and al, 0x04
  or cs:mouse_button_status, al

  sermouse_notfourth:

  sermouse_handler_end:
  pop cx
  pop dx
  jmp sermouse_handler_iret

  sermouse_handler_next:

  pop dx
  test word ptr cs:_old_serport_handler, 0xFFFF
  jnz sermouse_hdlr_jmp
  test word ptr cs:_old_serport_handler+2, 0xFFFF
  jnz sermouse_hdlr_jmp
  sermouse_handler_iret:
  mov al, 0x20
  out 0x20, al
  pop ax
  iret
  sermouse_hdlr_jmp:
  pop ax
  db 0xEA
  _old_serport_handler: dd 0
sermouse_handler_ endp

int33_handler_ proc far
  cmp ax, 0
  je return_driver_status
  cmp ax, 3
  je return_button_status
  cmp ax, 0x0B
  je return_motion

  jmp int33_end

  return_driver_status:
  dec ax
  mov bx, 3         ; 3 buttons
  jmp int33_end

  return_button_status:
  xor bh, bh
  mov bl, byte ptr cs:mouse_button_status
  and bl, 0x07
  mov cx, word ptr cs:cursor_col
  mov dx, word ptr cs:cursor_row
  jmp int33_end

  return_motion:
  xor cx, cx
  xchg cx, cs:mouse_dx
  xor dx, dx
  xchg dx, cs:mouse_dy

  int33_end:
  iret
int33_handler_ endp

tsr_end_:

no_ps2mouse_support:
mov ax, 1
ret

_resolution: db 2
_samplerate: db 4

try_ps2_load_ proc near

; reset mouse
mov ax, 0xC201
int 0x15
jc no_ps2mouse_support

; initialize
mov ax, 0xC205
mov bh, 3
int 0x15
jc no_ps2mouse_support

; set far call handler
mov ax, 0xC207
push cs
pop es
mov bx, offset ps2mouse_handler_
int 0x15
jc no_ps2mouse_support

; enable mouse
mov ax, 0xC200
mov bh, 1
int 0x15
jc no_ps2mouse_support

; set resolution
mov ax, 0xC203
mov bh, cs:_resolution
int 0x15
jc no_ps2mouse_support

; set sample rate
mov ax, 0xC202
mov bh, cs:_samplerate
int 0x15
jc no_ps2mouse_support

xor ax, ax
ret

try_ps2_load_ endp


END
