#define serialbase ((unsigned __far*)MK_FP(0x40, 0x00))

unsigned serport_count(void);
#pragma aux serport_count = \
  "int 0x11" \
  "mov cl, 9" \
  "shr ax, cl" \
  "and ax, 0x7" \
  value [ax] modify [cx];

void outbyte(unsigned port, unsigned byte);
#pragma aux outbyte = \
  "out dx, al" \
  parm [dx] [ax];

unsigned char inbyte(unsigned port);
#pragma aux inbyte = \
  "in al, dx" \
  parm [dx] value [al];

volatile unsigned char serial_recv_data;
unsigned detect_port;

void __far *old_serport_hdlr;

void __interrupt com_detect_intr(void)
{
  const unsigned port = detect_port;
  unsigned char c = inbyte(port+2);
  // was interrupt caused by this port && is data available?
  if(!(c & 0x01))
  {
    c = inbyte(port);
    if(!serial_recv_data)
      if(c == 'M' || c == '3' || c == 'Z')
        serial_recv_data = c;
  }
  else
    if(old_serport_hdlr)
      _chain_intr(old_serport_hdlr);

/*  __asm
  {
    mov ah, 0x0E
    mov al, c
    int 0x10
  }*/

  outbyte(0x20, 0x20);
}

int detect_serial_port(unsigned port)
{
  unsigned char intr = 0xC - (port & 1);
  unsigned char old_irq_mask;
  unsigned ioport = serialbase[port];
  unsigned char old_mcr = inbyte(ioport+4);
  unsigned char old_lcr = inbyte(ioport+3);
  unsigned old_div;
  unsigned char old_fifo = inbyte(ioport+2);

  printf("Probing serial mouse from COM%u\n", port+1);

  // enable DLAB
  outbyte(ioport+3, 0x80);
  // set baud divisor to 96 (baud rate 1200)
  old_div = inbyte(ioport) | (inbyte(ioport+1) << 8);
  outbyte(ioport, 96);
  outbyte(ioport+1, 0);
  // disable DLAB && set 7 data bits, no parity, 1 stop bit
  outbyte(ioport+3, 2);
  // disable FIFO
  outbyte(ioport+2, 0x00);
  // set interrupt when receiving
  outbyte(ioport+1, 0x01);

  // set RTS and DTR to low, IRQ disable
  outbyte(ioport+4, 0x00);
  // wait 200 ms so that the mouse has enough time to lose its power
  delay(200);
  // set DTR active
  outbyte(ioport+4, 0x01);
  // wait 100 ms so that the mouse has enough time to initialize itself
  delay(100);
  // set interrupt handlers
  old_serport_hdlr = _dos_getvect(intr);
  _dos_setvect(intr, com_detect_intr);
  // set IRQ mask
  old_irq_mask = inbyte(0x21);
  outbyte(0x21, old_irq_mask & ~(0x10 >> (port&1)));
  // set RTS (resets the mouse) and DTR to active, IRQ enable
  serial_recv_data = 0;
  detect_port = ioport;
  inbyte(ioport); // flush receive line?
  outbyte(ioport+4, 0xB);

  // wait for data ready
  delay(200);
  _dos_setvect(intr, old_serport_hdlr);

  if(serial_recv_data) // if a mouse was detected, return 1
  {
    return 1;
  }
  outbyte(0x21, old_irq_mask);
  outbyte(ioport+3, 0x80);
  outbyte(ioport, old_div);
  outbyte(ioport+1, old_div>>8);
  outbyte(ioport+3, old_lcr);
  outbyte(ioport+4, old_mcr);
  outbyte(ioport+2, old_fifo);
  return 0;
}

extern unsigned __based(__segname("_CODE")) sermouse_port;
void __interrupt sermouse_handler();
extern void __far * __based(__segname("_CODE")) old_serport_handler;

void set_serialmouse(unsigned port)
{
  unsigned char intr = 0x0C - (port & 1);
  sermouse_port = serialbase[port];
  old_serport_handler = _dos_getvect(intr);
  _dos_setvect(intr, sermouse_handler);
}

int detect_serial_mouse(void)
{
  unsigned n = serport_count();
  if(!n)
  {
    puts("No serial ports detected.");
    return 1;
  }
  printf("The system has %u serial ports.\n", n);

  while(n--)
  {
    if(detect_serial_port(n))
    {
      printf("Mouse detected in COM%u. Mouse type:\n", n+1);
      switch(serial_recv_data)
      {
        case 'M': puts("Normal"); break;
        case '3': puts("3-button"); break;
        case 'Z': puts("Mouse-wheel"); break;
      }
      set_serialmouse(n);
      return 0;
    }
    else
      puts("No mouse detected.");
  }
  return 1;
}
